"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const iwanthue_1 = __importDefault(require("iwanthue"));
const d3Chromatic = __importStar(require("d3-scale-chromatic"));
const shuffle_in_place_1 = require("pandemonium/shuffle-in-place");
const seedrandom_1 = __importDefault(require("seedrandom"));
const shapes_1 = require("./shapes");
function getDefaultDefaultValue(kind) {
    if (kind === 'color')
        return '#ccc';
    else
        return shapes_1.UNKNOWN_SHAPE;
}
class Palette {
    constructor(name, kind, map, defaultValue) {
        this.name = name;
        this.kind = kind;
        this.map = map;
        this.defaultValue = defaultValue;
        this.size = this.map.size;
    }
    get(key) {
        const value = this.map.get(key);
        if (value === undefined)
            return this.defaultValue;
        return value;
    }
    forEach(callback) {
        this.map.forEach(callback);
    }
    static getMacroDefault(kind) {
        if (kind === 'color')
            return '#ccc';
        else
            return shapes_1.UNKNOWN_SHAPE;
    }
    static fromScheme(name, kind, scheme, values, defaultValue) {
        const target = d3Chromatic['scheme' + scheme];
        let colors;
        if (Array.isArray(target[target.length - 1])) {
            const firstValidCount = target.findIndex((c) => Array.isArray(c));
            colors = target[Math.min(Math.max(firstValidCount, values.length), target.length)];
        }
        else {
            colors = target.slice();
            const shuffleInPlace = (0, shuffle_in_place_1.createShuffleInPlace)((0, seedrandom_1.default)(name));
            shuffleInPlace(colors);
            colors = colors.slice(0, Math.min(values.length, colors.length));
        }
        const map = new Map();
        values.slice(0, colors.length).forEach((v, i) => {
            map.set(v, colors[i]);
        });
        return new Palette(name, kind, map, defaultValue || getDefaultDefaultValue(kind));
    }
    static fromEntries(name, kind, entries, defaultValue) {
        return new Palette(name, kind, new Map(entries), defaultValue || getDefaultDefaultValue(kind));
    }
    static generateFromValues(name, kind, values, defaultValue) {
        if (kind === 'color') {
            const settings = {
                colorSpace: 'sensible',
                seed: name,
                clustering: 'force-vector',
                attempts: 5,
            };
            if (values.length === 0)
                return new Palette(name, kind, new Map(), defaultValue || getDefaultDefaultValue(kind));
            const colors = (0, iwanthue_1.default)(values.length, settings);
            const map = new Map();
            values.forEach((v, i) => {
                map.set(v, colors[i]);
            });
            return new Palette(name, kind, map, defaultValue || getDefaultDefaultValue(kind));
        }
        else {
            const map = new Map();
            shapes_1.USEFUL_SHAPES.slice(0, Math.min(values.length, shapes_1.USEFUL_SHAPES.length)).forEach((shape, i) => {
                map.set(values[i], shape.name);
            });
            return new Palette(name, kind, map, defaultValue || getDefaultDefaultValue(kind));
        }
    }
}
exports.default = Palette;
//# sourceMappingURL=palette.js.map