"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
// Taken and adapted from: https://github.com/jacomyal/sigma.js/blob/main/examples/png-snapshot/saveAsPNG.ts
const sigma_1 = __importDefault(require("sigma"));
function saveAsPNG(renderer, inputLayers) {
    const { width, height } = renderer.getDimensions();
    // This pixel ratio is here to deal with retina displays.
    // Indeed, for dimensions W and H, on a retina display, the canvases
    // dimensions actually are 2 * W and 2 * H. Sigma properly deals with it, but
    // we need to readapt here:
    const pixelRatio = window.devicePixelRatio || 1;
    const tmpRoot = document.createElement('DIV');
    tmpRoot.style.width = `${width}px`;
    tmpRoot.style.height = `${height}px`;
    tmpRoot.style.position = 'absolute';
    tmpRoot.style.right = '101%';
    tmpRoot.style.bottom = '101%';
    document.body.appendChild(tmpRoot);
    // Instantiate sigma:
    const tmpRenderer = new sigma_1.default(renderer.getGraph(), tmpRoot, renderer.getSettings());
    // Copy camera and force to render now, to avoid having to wait the schedule /
    // debounce frame:
    tmpRenderer.getCamera().setState(renderer.getCamera().getState());
    tmpRenderer.refresh();
    // Create a new canvas, on which the different layers will be drawn:
    const canvas = document.createElement('CANVAS');
    canvas.setAttribute('width', width * pixelRatio + '');
    canvas.setAttribute('height', height * pixelRatio + '');
    const ctx = canvas.getContext('2d');
    // Draw a white background first:
    ctx.fillStyle = '#fff';
    ctx.fillRect(0, 0, width * pixelRatio, height * pixelRatio);
    // For each layer, draw it on our canvas:
    const canvases = tmpRenderer.getCanvases();
    const layers = inputLayers
        ? inputLayers.filter((id) => !!canvases[id])
        : Object.keys(canvases);
    layers.forEach((id) => {
        ctx.drawImage(canvases[id], 0, 0, width * pixelRatio, height * pixelRatio, 0, 0, width * pixelRatio, height * pixelRatio);
    });
    // Save the canvas as a PNG image:
    const dataURL = canvas.toDataURL();
    // Cleanup:
    tmpRenderer.kill();
    tmpRoot.remove();
    return dataURL;
}
exports.default = saveAsPNG;
//# sourceMappingURL=saveAsPNG.js.map