"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.VisualVariableScalesBuilder = exports.CategorySummary = exports.AttributeCategories = exports.AttributeExtents = exports.Extent = void 0;
const multi_set_1 = __importDefault(require("mnemonist/multi-set"));
const palette_1 = __importDefault(require("./palette"));
const d3_scale_1 = require("d3-scale");
const d3Chromatic = __importStar(require("d3-scale-chromatic"));
/**
 * Constants.
 */
const MAX_CATEGORICAL_COLORS = 10;
const DEFAULT_DEFAULT_CONTINUOUS_NUMERICAL_VALUE = 1;
const DEFAULT_DEFAULT_CONTINUOUS_COLOR_VALUE = 'black';
const ESPILON = 1e-8;
/**
 * Helper functions.
 */
function isValidNumber(value) {
    return (typeof value === 'number' &&
        !Number.isNaN(value) &&
        value !== Infinity &&
        value !== -Infinity);
}
function rangeIsConstant(range) {
    if (!Array.isArray(range))
        return false;
    if (typeof range[0] === 'string')
        return range[0] === range[1];
    return Math.abs(range[0] - range[1]) <= ESPILON;
}
function getContinuousDefaultValue(variable) {
    if (typeof variable.range === 'string' ||
        typeof variable.range[0] === 'string') {
        // Color
        if (typeof variable.default === 'string')
            return variable.default;
        if (Array.isArray(variable.range) && typeof variable.range[0] === 'string')
            return variable.range[0];
        return DEFAULT_DEFAULT_CONTINUOUS_COLOR_VALUE;
    }
    else {
        // Size
        if (isValidNumber(variable.default))
            return variable.default;
        if (Array.isArray(variable.range) && isValidNumber(variable.range[0]))
            return variable.range[0];
        return DEFAULT_DEFAULT_CONTINUOUS_NUMERICAL_VALUE;
    }
}
function createContinuousScale(variable, extent) {
    const defaultValue = getContinuousDefaultValue(variable);
    if (rangeIsConstant(variable.range) || extent.isConstant()) {
        return () => defaultValue;
    }
    let scale;
    const definition = variable.scale;
    let isLogPlusOne = false;
    if (!definition)
        scale = (0, d3_scale_1.scaleLinear)();
    else {
        const [type, param] = definition;
        if (type === 'lin')
            scale = (0, d3_scale_1.scaleLinear)();
        else if (type === 'log' || type === 'log+1') {
            isLogPlusOne = type === 'log+1';
            scale = (0, d3_scale_1.scaleLog)();
            if (param)
                scale.base(param);
        }
        else if (type === 'pow') {
            scale = (0, d3_scale_1.scalePow)();
            scale.exponent(param ? param : 2);
        }
        else if (type === 'sqrt') {
            scale = (0, d3_scale_1.scaleSqrt)();
            if (param)
                scale.exponent(1 / param);
        }
        else
            throw new Error('unknown scale type');
    }
    const offset = isLogPlusOne ? 1 : 0;
    scale.domain([extent.min + offset, extent.max + offset]);
    if (typeof variable.range === 'string') {
        const chromatic = d3Chromatic['interpolate' + variable.range];
        return (attr) => {
            let value = attr[variable.attribute];
            if (!isValidNumber(value))
                return defaultValue;
            value += offset;
            return chromatic(scale(value));
        };
    }
    else if (variable.range) {
        scale.range(variable.range);
        return (attr) => {
            let value = attr[variable.attribute];
            if (!isValidNumber(value))
                return defaultValue;
            value += offset;
            return scale(value);
        };
    }
    throw new Error('could not create valid continuous scale');
}
/**
 * Helper classes.
 */
class Extent {
    constructor() {
        this.min = Infinity;
        this.max = -Infinity;
    }
    add(value) {
        if (value < this.min)
            this.min = value;
        if (value > this.max)
            this.max = value;
    }
    isConstant() {
        return this.min === Infinity || this.min === this.max;
    }
}
exports.Extent = Extent;
class AttributeExtents {
    constructor(names) {
        this.attributes = {};
        // NOTE: this naturally deduplicates names
        names.forEach((name) => (this.attributes[name] = new Extent()));
    }
    add(attributes) {
        for (const name in this.attributes) {
            const value = attributes[name];
            if (!isValidNumber(value))
                continue;
            this.attributes[name].add(value);
        }
    }
}
exports.AttributeExtents = AttributeExtents;
class AttributeCategories {
    constructor(names) {
        this.attributes = {};
        // NOTE: this naturally deduplicates names
        names.forEach((name) => (this.attributes[name] = new multi_set_1.default()));
    }
    add(attributes) {
        for (const name in this.attributes) {
            this.attributes[name].add(attributes[name]);
        }
    }
}
exports.AttributeCategories = AttributeCategories;
class CategorySummary {
    constructor(name, kind, palette, overflowing = false) {
        this.name = name;
        this.kind = kind;
        this.palette = palette;
        this.overflowing = overflowing;
    }
    static fromTopValues(name, kind, frequencies, defaultValue, scheme, maxCount = MAX_CATEGORICAL_COLORS) {
        const count = Math.min(maxCount, frequencies.dimension);
        const topValues = frequencies.top(count);
        const overflowing = count < frequencies.dimension;
        const values = topValues.map((item) => item[0]);
        const palette = !scheme || scheme === 'IWantHue'
            ? palette_1.default.generateFromValues(name, kind, values, defaultValue)
            : palette_1.default.fromScheme(name, kind, scheme, values, defaultValue);
        return new CategorySummary(name, kind, palette, overflowing);
    }
    static fromEntries(name, kind, entries, defaultValue) {
        const palette = palette_1.default.fromEntries(name, kind, entries, defaultValue);
        return new CategorySummary(name, kind, palette);
    }
}
exports.CategorySummary = CategorySummary;
class VisualVariableScalesBuilder {
    constructor(visualVariables, maxCategories = MAX_CATEGORICAL_COLORS) {
        this.variables = visualVariables;
        this.maxCategories = maxCategories;
        const nodeExtentAttributes = [];
        const nodeCategoryAttributes = [];
        const edgeExtentAttributes = [];
        const edgeCategoryAttributes = [];
        for (const variableName in visualVariables) {
            const variable = visualVariables[variableName];
            if (variableName.startsWith('node')) {
                if (variable.type === 'category') {
                    if (!variable.palette || typeof variable.palette === 'string')
                        nodeCategoryAttributes.push(variable.attribute);
                }
                else if (variable.type === 'continuous' ||
                    variableName === 'nodeSize') {
                    nodeExtentAttributes.push(variable.attribute);
                }
            }
            else if (variableName.startsWith('edge')) {
                if (variable.type === 'category') {
                    if (!variable.palette || typeof variable.palette === 'string')
                        edgeCategoryAttributes.push(variable.attribute);
                }
                else if (variable.type === 'continuous') {
                    edgeExtentAttributes.push(variable.attribute);
                }
            }
        }
        this.nodeExtents = new AttributeExtents(nodeExtentAttributes);
        this.edgeExtents = new AttributeExtents(edgeExtentAttributes);
        this.nodeCategories = new AttributeCategories(nodeCategoryAttributes);
        this.edgeCategories = new AttributeCategories(edgeCategoryAttributes);
    }
    readGraph(graph) {
        graph.forEachNode((node, attr) => {
            this.nodeExtents.add(attr);
            this.nodeCategories.add(attr);
        });
        graph.forEachEdge((edge, attr) => {
            this.edgeExtents.add(attr);
            this.edgeCategories.add(attr);
        });
    }
    build() {
        const scales = {};
        for (const variableName in this.variables) {
            const variable = this.variables[variableName];
            let scale = null;
            // Raw variables
            if (variable.type === 'raw') {
                scale = (attr) => attr[variable.attribute] || variable.default;
            }
            // Constant variables
            else if (variable.type === 'constant') {
                scale = () => variable.default;
            }
            // Category variables
            else if (variable.type === 'category') {
                const categories = variableName.startsWith('node')
                    ? this.nodeCategories
                    : this.edgeCategories;
                const summary = variable.palette && typeof variable.palette !== 'string'
                    ? CategorySummary.fromEntries(variable.attribute, variable.kind || 'color', variable.palette, variable.default)
                    : CategorySummary.fromTopValues(variable.attribute, variable.kind || 'color', categories.attributes[variable.attribute], variable.default, variable.palette, this.maxCategories);
                const palette = summary.palette;
                scale = (attr) => palette.get(attr[variable.attribute]);
                scale.summary = summary;
            }
            // Continuous variables
            else if (variable.type === 'continuous') {
                const extents = variableName.startsWith('node')
                    ? this.nodeExtents
                    : this.edgeExtents;
                const extent = extents.attributes[variable.attribute];
                scale = createContinuousScale(variable, extent);
            }
            if (scale)
                scales[variableName] = scale;
        }
        return scales;
    }
    inferLabelRenderedSizeThreshold() {
        const variable = this.variables.nodeSize;
        if (variable.type === 'continuous') {
            const range = variable.range;
            return Math.min(6, range[0]);
        }
        return 6;
    }
}
exports.VisualVariableScalesBuilder = VisualVariableScalesBuilder;
//# sourceMappingURL=visual-variables.js.map